/*
         o     o         o
         |     |        /
         |     |       /
  o======o=====o======o=======o
        / \            \
       /   \            \
      o     o            o

Small Wide World == Open Source Graph Optimization with an editor
Copyright (C) 2015-2016 Javantea
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Code in this file is based on SWW Minimize 1
by Javantea
Sept 30, 2015-Jan 25, 2016
Based on N Atom molecule using Basin-Hopping Method (aka L-BFGS-B)
by Javantea
July 23, 2014
Based on metropolis_min1.py by Javantea, July 17-22, 2014
Based on s_4atom.py by Javantea, June 24, 2014

*/

/*
 * LJ_Potential(r0=40, scale_factor=0.055)
	Proper Lennard-Jones 6 12 Potential.
 */
function LJ_Potential(r0, scale_factor)
{
	if (r0 == undefined) { r0 = 40; }
	if (scale_factor == undefined) { scale_factor = 0.055; }
	this.r0 = r0;
	this.scale_factor = scale_factor;
	this.A = 2 * Math.pow(r0, 6) * scale_factor;
	this.B = Math.pow(r0, 12) * scale_factor;
} //#end function __init__()
// This is a class.
new LJ_Potential();

/*
 * LJ_Potential.V_LJ(r)
	Attempt at a Lennard-Jones 6 12 potential
	The potential is very high when the two nodes are close.
	The potential is low when two nodes are far away.
	The potential is lowest when two nodes are at the r0 distance, not too 
	close, not too far away.
	Non-Connected nodes should have a LJ potential.
	r0 is the value for which this LJ potential is optimized.
 */
LJ_Potential.prototype.V_LJ = function(r) {
	// asm.js...
	r = +r;
	//# This is a divide by zero protection.
	if (r < 1e-3) {
		return this.V_LJ(1e-3);
	}//#end if
	var r6 = 1 / Math.pow(r, 6);
	return ((-this.A*r6) + (this.B*r6*r6));
} //#end function V_LJ(r)

/*
 * LJ_Potential.V_LJ_fast(r)
 * This function is equivalent but faster than V_LJ(r.length())
 */
LJ_Potential.prototype.V_LJ_fast = function(r) {
	// asm.js...
	var lsq = +r.lensq();
	//# This is a divide by zero protection.
	if (lsq < 1e-6) {
		return this.V_LJ(1e-3);
	}//#end if
	var r6 = 1 / Math.pow(lsq, 3);
	return ((-this.A*r6) + (this.B*r6*r6));
} //#end function V_LJ(r)
//#end class LJ_Potential

/*
 * V_Harmonic(r, b0)
	Attempt at a Harmonic potential
	Connected nodes should have a harmonic potential.
	b0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
 */
function V_Harmonic(r, b0) {
	// asm.js...
	r = +r;
	b0 = +b0;
	var A = +1;
	return A * Math.pow(r - b0, 2);
} //#end function V_Harmonic(r, b0)

/*
 * V_Harmonic_Angle(theta, theta0)
	Attempt at a Harmonic potential for the angle between two connected items
	Connected nodes should have a harmonic potential.
	theta0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
 */
function V_Harmonic_Angle(theta, theta0) {
	// asm.js...
	theta = +theta;
	theta0 = +theta0;
	var A = +2;
	return A * Math.pow(theta - theta0, 2);
} //#end function V_Harmonic_Angle(theta, theta0)

/*
 * V_SmallWideWorld(x, conns, b0, lj_conf, nonConns)
	Just like a molecule but in 2D.
 */
function V_SmallWideWorld(x, conns, b0, lj_conf, nonConns) {
	// TODO: asm.js...
	//debug('V_SWW(' + JSON.stringify(x) + ', ...)');
	var Potential = +0.0;
	var i;
	var k;
	var jindex;
	var pt_i, pt_j;
	var dps_all = new Array(conns.length).fill(null);
	for (i = 0; i < conns.length; i++) {
		dps_all[i] = [];
	}
	// Generate conns list for each node
	for (i = 0; i < conns.length; i++) {
		pt_i = new Vector2(x[2*i], x[2*i+1]);
		for (jindex = 0; jindex < conns[i].length; jindex++) {
			var j = conns[i][jindex];
			pt_j = new Vector2(x[2*j], x[2*j+1]);
			var dp = pt_i.sub(pt_j);
			dps_all[i].push(dp);
			dps_all[j].push(dp.mul(-1));
		}
	}
	//debug("verts: " + JSON.stringify(verts_all))
	//debug("dps: " + JSON.stringify(dps_all));
	//# Harmonic potential with length b0.
	for (i = 0; i < conns.length; i++) {
		pt_i = new Vector2(x[2*i], x[2*i+1]);
		var dps = dps_all[i];
		for (jindex = 0; jindex < conns[i].length; jindex++) {
			var j = conns[i][jindex];
			pt_j = new Vector2(x[2*j], x[2*j+1]);
			var dp = pt_i.sub(pt_j);
			var dp_len = dp.length();
			Potential += V_Harmonic(dp_len, b0);
		} //#next j
		if (dps.length >= 2) {
			//# Angular Harmonic calculation
			var theta0 = 2 * Math.PI / dps.length;
			//# theta is computed for axb, bxc, cxa
			var thetas = [];
			for (k = 0; k < dps.length; k++) {
				var ax = dps[k];
				thetas.push(Number_mod(Math.atan2(ax.y, ax.x), 2*Math.PI));
			}
			thetas.sort(function(a, b) { return a - b; });
			for (k = 0; k < thetas.length; k++) {
				var b = (k+1) % thetas.length;
				var theta = Number_mod(thetas[b] - thetas[k], 2*Math.PI);
				Potential += V_Harmonic_Angle(theta, theta0);
			} //#next i
		} //#end if
	} //#next i
	//# Lennard-Jones for non-connected nodes.
	for (i = 0; i < nonConns.length; i++) {
		pt_i = new Vector2(x[2*i], x[2*i+1]);
		for (jindex = 0; jindex < nonConns[i].length; jindex++) {
			var j = nonConns[i][jindex];
			pt_j = new Vector2(x[2*j], x[2*j+1]);
			var dp = pt_i.sub(pt_j);
			Potential += lj_conf.V_LJ_fast(dp);
		} //#next j
	} //#next i
	//debug('== ' + Potential);
	return Potential;
} //#end function V_SmallWideWorld(x, conns, b0, lj_conf)

/*
 * V_SmallWideWorld_verbose(x, conns, b0, lj_conf, nonConns)
	Just like a molecule but in 2D.
 */
function V_SmallWideWorld_verbose(x, conns, b0, lj_conf, nonConns) {
	// TODO: asm.js...
	var Potential = +0.0;
	var HP = +0.0;
	var HPA = +0.0;
	var i;
	var k;
	var jindex;
	var pt_i, pt_j;
	var dps_all = new Array(conns.length).fill(null);
	for (i = 0; i < conns.length; i++) {
		dps_all[i] = [];
	}
	// Generate conns list for each node
	for (i = 0; i < conns.length; i++) {
		pt_i = new Vector2(x[2*i], x[2*i+1]);
		for (jindex = 0; jindex < conns[i].length; jindex++) {
			var j = conns[i][jindex];
			pt_j = new Vector2(x[2*j], x[2*j+1]);
			var dp = pt_i.sub(pt_j);
			dps_all[i].push(dp);
			dps_all[j].push(dp.mul(-1));
		}
	}
	debug("dps: " + JSON.stringify(dps_all));
	//# Harmonic potential with length b0.
	for (i = 0; i < conns.length; i++) {
		debug('i: ' + i);
		pt_i = new Vector2(x[2*i], x[2*i+1]);
		var dps = dps_all[i];
		for (jindex = 0; jindex < conns[i].length; jindex++) {
			var j = conns[i][jindex];
			pt_j = new Vector2(x[2*j], x[2*j+1]);
			var dp = pt_i.sub(pt_j);
			var dp_len = dp.length();
			var v = V_Harmonic(dp_len, b0);
			Potential += v;
			HP += v;
			debug("r: " + dp_len + ", V: " + v);
		} //#next j
		if (dps.length >= 2) {
			//# Angular Harmonic calculation
			theta0 = 2 * Math.PI / dps.length;
			//# theta is computed for axb, bxc, cxa
			var thetas = [];
			for (k = 0; k < dps.length; k++) {
				var ax = dps[k];
				thetas.push(Number_mod(Math.atan2(ax.y, ax.x), 2*Math.PI));
			}
			// In order to get numerical sort, we have to do this.
			thetas.sort(function(a, b) { return a - b; });
			debug("thetas: " + thetas);
			for (k = 0; k < thetas.length; k++) {
				var b = (k+1) % thetas.length;
				var theta = Number_mod(thetas[b] - thetas[k], 2*Math.PI);
				v = V_Harmonic_Angle(theta, theta0);
				Potential += v;
				HPA += v;
				debug("theta: " + theta + ", theta0: " + theta0 + ", V: " + v);
			} //#next i
		} //#end if
	} //#next i
	debug("Harmonic Potential: " + (HP));
	debug("Angular Harmonic Potential: " + (HPA));
	//# Lennard-Jones for non-connected nodes.
	for (i = 0; i < nonConns.length; i++) {
		pt_i = new Vector2(x[2*i], x[2*i+1]);
		for (jindex = 0; jindex < nonConns[i].length; jindex++) {
			var j = nonConns[i][jindex];
			pt_j = new Vector2(x[2*j], x[2*j+1]);
			var dp = pt_i.sub(pt_j);
			v = lj_conf.V_LJ_fast(dp);
			Potential += v;
			debug("r: " + dp_len + ", V: " + v);
		} //#next j
	} //#next i
	debug("LJ Potential: " + (Potential-(HP+HPA)));
	return Potential;
} //#end function V_SmallWideWorld_verbose(x, conns, b0, lj_conf)

function computeV_SmallWideWorld(map1, b0)
{
	if (b0 == undefined) { b0 = +40; }
	var lj_conf = new LJ_Potential();
	//# The Metropolis algorithm in a nutshell.
	var x0conns = map1.getX0Conns(true);
	var x0 = x0conns[0];
	var conns = x0conns[1];
	var nonConns = getNonConns(conns, true);
	var V0 = V_SmallWideWorld(x0, conns, b0, lj_conf, nonConns);
	debug("Potential: " + V0);
	return V0;
}

/**
 * LBFGSB(map1, b0=40, gtol=1e-2)
 * gtol is the gradient tolerance.
 * 1e-5 is the default but 1e-2 is faster and not a lot worse. Up to you.
 * Returns the Potential found by the algorithm.
 * Sets the node positions to the minimum, if smaller than the initial potential.
 */
function LBFGSB(map1, b0, gtol)
{
	if (b0 == undefined) { b0 = +40; }
	if (gtol == undefined) { gtol = +1e-2; }
	var lj_conf = new LJ_Potential();
	//# The Metropolis algorithm in a nutshell.
	var x0conns = map1.getX0Conns(true);
	var x0 = x0conns[0];
	var conns = x0conns[1];
	var nonConns = getNonConns(conns, true);
	var V0 = V_SmallWideWorld(x0, conns, b0, lj_conf, nonConns);
	debug("Initial Potential: " + V0);
	// This is amazing, this this the minimization.
	var res = _minimize_lbfgsb(V_SmallWideWorld, x0, [conns, b0, lj_conf, nonConns], null, null, null, 10, 2.2204460492503131e-09, gtol);
	debug(JSON.stringify(res));
	if (res.success == false) {
		debug("Minimization failed.");
	}
	var Vmin = V_SmallWideWorld(res.x, conns, b0, lj_conf, nonConns);
	debug("Minimized Potential: " + Vmin);
	if(Vmin < V0) {
		// res.x is a 1d array, we want an array of 2d vectors.
		var x_min = reshape(Array.from(res.x), map1.points.length, 2);
		// Now we set using setPos.
		map1.setPos(x_min);
	}
	return Vmin;
}
