#!/usr/bin/env python3
"""
Packing Unconnected Nodes According to Lennard-Jones 6-12 Potential
by Javantea
Feb 11, 2016

Given a set of unconnected nodes in 2D, we can easily pack them according to minimization. First, let's look at local minimization results.
"""
from __future__ import print_function
import sys
import sww_optimize1
import numpy as np
from scipy.optimize import minimize
mapmake1 = sww_optimize1.mapmake1
ConnList = sww_optimize1.ConnList

def main():
	nodes = 20
	# Skip minimization, try out our algorithm.
	mini1 = False
	if len(sys.argv) > 1:
		nodes = int(sys.argv[1])
	#end if
	base_filename = 'unconn_{0}.json'.format(nodes)

	map1 = mapmake1.Map()
	map1.addPoints([mapmake1.Point('p_'+str(i)) for i in range(nodes)])
	map1.sortLinear()
	map1.centerBound()
	open(base_filename,'w').write(map1.saveJson())
	# It's important for it not to be linear if it's going to be minimized
	# because linear is a sort of ridge.
	if mini1:
		map1.randomize(8)
	#end if

	# Found a bug in sortLinear!

	LJ_Potential = sww_optimize1.LJ_Potential
	V_SmallWideWorld = sww_optimize1.V_SmallWideWorld
	getNonConns = mapmake1.getNonConns
	x0, conns = map1.getX0Conns(True)
	conns_cpp = ConnList(conns)
	nonConns = ConnList(getNonConns(conns, True))
	b0 = 40
	lj_conf = LJ_Potential()
	if mini1:
		minimizer_opts = {"method": "L-BFGS-B", 'args':(conns_cpp, b0, lj_conf, nonConns), 'options':{'gtol':1e-2, 'eps':1e-4}}

		# TODO: Use a potential with LJ only
		res = minimize(V_SmallWideWorld, x0, **minimizer_opts)
		print(res)
		Vmin = V_SmallWideWorld(res.x, conns_cpp, b0, lj_conf, nonConns)
		print('Local Minimized:', Vmin)
		m = np.matrixlib.matrix(res.x)
		m = m.reshape(len(map1.points), 2)
		map1.setPos([np.array(xa)[0] for xa in m])
		map1.centerBound()
		open(base_filename.replace('.json', 'min.json'), 'w').write(map1.saveJson())
		open(base_filename.replace('.json', 'min.svg'), 'w').write(map1.drawSvg())
	#end if
	"""
	Impressively, the global optimizer found a better configuration which proves its value.
	python3 ../sww_optimize1.py -f sww_optimize1out20.json 20 0 10
	basinhopping step 0: f -4.8427
	basinhopping step 1: f -4.8427 trial_f -4.60915 accepted 0  lowest_f -4.8427
	basinhopping step 2: f -4.8427 trial_f -4.6241 accepted 0  lowest_f -4.8427
	basinhopping step 3: f -4.87726 trial_f -4.87726 accepted 1  lowest_f -4.87726
	found new global minimum on step 3 with function value -4.87726
	basinhopping step 4: f -4.99962 trial_f -4.99962 accepted 1  lowest_f -4.99962
	found new global minimum on step 4 with function value -4.99962
	basinhopping step 5: f -4.99962 trial_f -4.50361 accepted 0  lowest_f -4.99962
	basinhopping step 6: f -4.75848 trial_f -4.75848 accepted 1  lowest_f -4.99962
	basinhopping step 7: f -4.86859 trial_f -4.86859 accepted 1  lowest_f -4.99962
	basinhopping step 8: f -5.00033 trial_f -5.00033 accepted 1  lowest_f -5.00033
	found new global minimum on step 8 with function value -5.00033
	basinhopping step 9: f -4.99347 trial_f -4.99347 accepted 1  lowest_f -5.00033
	basinhopping step 10: f -5.11088 trial_f -5.11088 accepted 1  lowest_f -5.11088
	found new global minimum on step 10 with function value -5.11088
	sww_optimize1out22.svg
	sww_optimize1out22.json
	Configuration:
	[ 365.2515838   207.47207467  320.11648196  337.46259412  335.17986084
	233.50327175  282.51263815  142.53740588  327.62079052  194.58759743
	290.12125344  181.6669493   297.66580944  220.58819562  267.59349495
	246.67140931  350.16971534  311.38963921  275.13409811  285.73423761
	305.21281978  259.47376811  320.10498226  155.52583773  387.55739989
	324.74771716  312.69409172  298.39264031  380.36876623  285.4607241
	357.71578772  168.43418358  372.81903025  246.44665214  252.44260011
	168.65320411  342.72574586  272.41653113  260.03565323  207.69129735]
	Minimum potential: -5.110883789255946
	Distance: 137.6034613091832
	Initial Potential: -4.842700554356694
	Initial Distance: 39.78595857837286

	So the correct configuration for 20 unconnected nodes should have lower potential than -5.11 or equal.
	My guess is that it's using a simple eqilateral triangle shape for any 3 nearby nodes. So we can do that with simple trig.
	Make it a square.

	"""
	# This function was written here, but now it's a part of mapmake1.py.
	map1.sortLattice(b0)
	map1.centerBound()
	x1, conns = map1.getX0Conns(True)
	Vmin = V_SmallWideWorld(x1, conns_cpp, b0, lj_conf, nonConns)
	print('Energy:', Vmin)
	open(base_filename.replace('.json', 'lattice1.json'), 'w').write(map1.saveJson())
	"""
	This lattice is very slightly worse than the optimized lattice, which says a lot about the quality of my algorithm. It took 3 steps for basinhopping to find a -5.10004 local minimum.
	"""
#end def main()

if __name__ == '__main__':
	main()
#end if
